/*
 * Copyright (c) 2013-2014 by Farsight Security, Inc.
 * Copyright (c) 2023 by the University of Southern California.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dnstapmq

import (
	"bytes"
	"errors"
	"fmt"
	"net"

	"github.com/miekg/dns"
)

var msgid = 1

func bool_to_string(b bool) string {
	if (b) {
		return "1"
	} else {
		return "0"
	}
}

//func mqConvertMessageSection(msg *Msg, prefix_s *bytes.Buffer, s *bytes.Buffer) {
//	// xxx: id qr opcode aa tc rd ra z ad cd rcode qdcount ancount nscount arcount edns_present edns_udp_size edns_extended_rcode edns_version edns_z msglen ipttl name type class
//
//	// foreach rr
//	
//	if m.QueryZone != nil {
//		name, _, err := dns.UnpackDomainName(m.QueryZone, 0)
//		if err != nil {
//			s.WriteString("-\t")
//		} else {
//			s.WriteString(fmt.Sprint(name, "\n"))
//		}
//	}
//
//	if m.QueryMessage != nil {
//		msg := new(dns.Msg)
//		err := msg.Unpack(m.QueryMessage)
//		if err != nil {
//			fmt.Fprintf(s, "  # query_message: parse failed: %v\n", err)
//		} else {
//			s.WriteString("  query_message: |\n")
//			s.WriteString("    " + strings.Replace(strings.TrimSpace(msg.String()), "\n", "\n    ", -1) + "\n")
//		}
//	}
//	if m.ResponseMessage != nil {
//		msg := new(dns.Msg)
//		err := msg.Unpack(m.ResponseMessage)
//		if err != nil {
//			fmt.Fprintf(s, "  # response_message: parse failed: %v\n", err)
//		} else {
//			s.WriteString("  response_message: |\n")
//			s.WriteString("    " + strings.Replace(strings.TrimSpace(msg.String()), "\n", "\n    ", -1) + "\n")
//		}
//	}
//	s.WriteString("---\n")
//}

func mqConvertMessage(m *Message, s *bytes.Buffer) error {
	var thisMsgId = msgid
	msgid++

	// If we did multiple RRs, we would build a prefix_s for each record.
	// We only do one query here, for message_question, so prefix_s is an alias.
	var prefix_s = s
	prefix_s.WriteString(fmt.Sprint(thisMsgId, "\t"))
	
	var timeSec = m.QueryTimeSec
	var timeNsec = m.QueryTimeNsec
	if m.ResponseTimeSec != nil {
		timeSec = m.ResponseTimeSec
		if m.ResponseTimeNsec != nil {
			timeNsec = m.ResponseTimeNsec
		} else {
			*timeNsec = 0
		}
	}
	if timeNsec == nil {
		*timeNsec = 0
	}
	if timeSec != nil {
		prefix_s.WriteString(fmt.Sprintf("%d.%09d\t", *timeSec, *timeNsec))
	} else {
		prefix_s.WriteString("-\t")
	}
	
	if m.QueryAddress != nil {
		prefix_s.WriteString(fmt.Sprint(net.IP(m.QueryAddress), "\t"))
	} else {
		prefix_s.WriteString("-\t")
	}
	if m.QueryPort != nil {
		prefix_s.WriteString(fmt.Sprint(*m.QueryPort, "\t"))
	} else {
		prefix_s.WriteString("-\t")
	}


	if m.ResponseAddress != nil {
		prefix_s.WriteString(fmt.Sprint(net.IP(m.ResponseAddress), "\t"))
	} else {
		prefix_s.WriteString("-\t")
	}
	if m.ResponsePort != nil {
		prefix_s.WriteString(fmt.Sprint(*m.ResponsePort, "\t"))
	} else {
		prefix_s.WriteString("-\t")
	}

	if m.SocketProtocol != nil {
		// xxx: convert to lower case, check tls
		prefix_s.WriteString(fmt.Sprint(m.SocketProtocol, "\t"))
	}

	//
	// now unpack the actualy dns message
	//
	msg := new(dns.Msg)
	if m.ResponseMessage != nil {
		err := msg.Unpack(m.ResponseMessage)
		if err != nil {
			return errors.New("cannot unpack reponse")
		}
	} else if m.QueryMessage != nil {
		err := msg.Unpack(m.QueryMessage)
		if err != nil {
			return errors.New("cannot unpack query")
		}
	} else {
		// no message!
		return errors.New("no message")
	}

	//
	// here's where we need to iterate over the RRs
	//
	// BUT: as a simplifying assumption, with
	// message_question format we only do the query,
	// for both query and response.
	// (Dnsanon has other options to do other RRsets,
	// but we don't use that in RSSAC002 processing.)
	//
	
	// xxx: id qr opcode aa tc rd ra z ad cd rcode qdcount ancount nscount arcount edns_present edns_udp_size edns_extended_rcode edns_version edns_z msglen ipttl name type class
	s.WriteString(fmt.Sprint(msg.Id, "\t", 
		bool_to_string(msg.Response), "\t",
		msg.Opcode, "\t",
		bool_to_string(msg.Authoritative), "\t",
		bool_to_string(msg.Truncated), "\t",
		bool_to_string(msg.RecursionDesired), "\t",
		bool_to_string(msg.RecursionAvailable), "\t",
		bool_to_string(msg.Zero), "\t",
		bool_to_string(msg.AuthenticatedData), "\t",
		bool_to_string(msg.CheckingDisabled), "\t",
		msg.Rcode, "\t",
		len(msg.Question), "\t",
		len(msg.Answer), "\t",
		len(msg.Ns), "\t",
		len(msg.Extra), "\t"))
	// xxx: edns_present edns_udp_size edns_extended_rcode edns_version edns_z msglen ipttl name type class
	if msg.IsEdns0() != nil {
		opt := msg.IsEdns0()
		// xxx: do we need to add in extended rcode?
		s.WriteString(fmt.Sprint("1\t",
			int(opt.UDPSize()), "\t",
			opt.ExtendedRcode(), "\t",
			opt.Version(), "\t",
			"-\t")) // no edns_z
	} else {
		s.WriteString("0\t-\t-\t-\t-\t")
	}
	// xxx: msglen ipttl name type class
	var q = msg.Question[0]
	s.WriteString(fmt.Sprint(msg.Len(), "\t",
		"-\t",
		q.Name, "\t",
		dns.Type(q.Qtype).String(), "\t",
		dns.Class(q.Qclass).String(), "\n"))
		
	return nil
}

var mQHeaderWritten = false

// MQFormat renders a dnstap message in MQ format.
func MQFormat(dt *Dnstap) (out []byte, ok bool) {
	var s bytes.Buffer

	if !mQHeaderWritten {
		s.WriteString("#fsdb -F t msgid time srcip srcport dstip dstport protocol id qr opcode aa tc rd ra z ad cd rcode qdcount ancount nscount arcount edns_present edns_udp_size edns_extended_rcode edns_version edns_z msglen ipttl name type class\n")
		mQHeaderWritten = true
	}

	// dt.Identity
	// dt.Version
	if *dt.Type == Dnstap_MESSAGE {
		err := mqConvertMessage(dt.Message, &s)
		if err != nil {
			return s.Bytes(), false
		}
	}
	return s.Bytes(), true
}
